@@ -1,34 +1,57 @@
 /*
  *	file.c
  *
- *	File Access Module
+ *	file accessing module
  */
 
 #include <fcntl.h>
 #include "fd.h"
+#include "termio.h"
 #include "func.h"
 #include "kanji.h"
 
-#ifndef	ENOSPC
-#define	ENOSPC	EACCES
-#endif
-
 #if	MSDOS
 #include <process.h>
-extern int getcurdrv __P_((VOID_A));
-# ifndef	_NOUSELFN
-extern char *preparefile __P_((char *, char *));
-# endif
-#else
-#include <pwd.h>
-#include <grp.h>
 #endif
 
-#ifndef	_NODOSDRIVE
-extern int lastdrv;
-# if	MSDOS && !defined (_NOUSELFN)
-extern int checkdrive __P_((int));
+#define	MAXTMPNAMLEN	8
+#define	LOCKEXT		"LCK"
+#ifdef	_NODOSDRIVE
+#define	DOSDIRENT	32
+#define	LFNENTSIZ	13
+#endif
+#define	LFNONLY		" +,;=[]"
+#define	DOSBODYLEN	8
+#define	DOSEXTLEN	3
+
+#ifndef	O_BINARY
+#define	O_BINARY	0
+#endif
+#ifndef	O_ACCMODE
+#define	O_ACCMODE	(O_RDONLY | O_WRONLY | O_RDWR)
+#endif
+#ifndef	ENOSPC
+#define	ENOSPC		EACCES
+#endif
+#ifndef	EIO
+#define	EIO		ENODEV
+#endif
+#ifndef	ETIMEDOUT
+#define	ETIMEDOUT	EIO
+#endif
+#ifndef	L_SET
+# ifdef	SEEK_SET
+# define	L_SET	SEEK_SET
+# else
+# define	L_SET	0
 # endif
+#endif	/* !L_SET */
+
+#if	MSDOS
+extern int getcurdrv __P_((VOID_A));
+#endif
+#if	MSDOS && !defined (_NOUSELFN) && !defined (_NODOSDRIVE)
+extern int checkdrive __P_((int));
 #endif
 
 extern int mark;
@@ -40,50 +63,36 @@ extern int noconv;
 extern int dosdrive;
 #endif
 
-#define	MAXTMPNAMLEN	8
-#ifndef	O_BINARY
-#define	O_BINARY	0
+#ifndef	NOFLOCK
+static int NEAR fcntllock __P_((int, int));
 #endif
-#ifdef	_NODOSDRIVE
-#define	DOSDIRENT	32
-#define	LFNENTSIZ	13
-#endif
-#define	LFNONLY		" +,;=[]"
-#define	DOSBODYLEN	8
-#define	DOSEXTLEN	3
-
-#ifndef	L_SET
-# ifdef	SEEK_SET
-# define	L_SET	SEEK_SET
-# else
-# define	L_SET	0
-# endif
-#endif
-
+static char *NEAR excllock __P_((CONST char *, int));
 #ifdef	_NODOSDRIVE
 #define	nodoschdir	Xchdir
 #define	nodosgetwd	Xgetwd
 #define	nodosmkdir	Xmkdir
 #define	nodosrmdir	Xrmdir
 #else
-static int NEAR nodoschdir __P_((char *));
+static int NEAR nodoschdir __P_((CONST char *));
 static char *NEAR nodosgetwd __P_((char *));
-static int NEAR nodosmkdir __P_((char *, int));
-static int NEAR nodosrmdir __P_((char *));
+static int NEAR nodosmkdir __P_((CONST char *, int));
+static int NEAR nodosrmdir __P_((CONST char *));
 #endif
-static int NEAR cpfile __P_((char *, char *, struct stat *, struct stat *));
+static int NEAR cpfile __P_((CONST char *, CONST char *,
+		struct stat *, struct stat *));
 static VOID changemes __P_((VOID_A));
 static int NEAR genrand __P_((int));
-static int dounlink __P_((char *));
+static int dounlink __P_((CONST char *));
+static int dormdir __P_((CONST char *));
 #ifndef	_NOWRITEFS
-static int NEAR isexist __P_((char *));
-static int NEAR realdirsiz __P_((char *, int, int, int, int));
+static int NEAR isexist __P_((CONST char *));
+static int NEAR realdirsiz __P_((CONST char *, int, int, int, int));
 static int NEAR getnamlen __P_((int, int, int, int, int));
-static int NEAR saferename __P_((char *, char *));
-static char *NEAR maketmpfile __P_((int, int, char *, char *));
+static int NEAR saferename __P_((CONST char *, CONST char *));
+static char *NEAR maketmpfile __P_((int, int, CONST char *, CONST char *));
 #if	!MSDOS
-static off_t NEAR getdirblocksize __P_((char *));
-static u_char *NEAR getentnum __P_((char *, off_t));
+static off_t NEAR getdirblocksize __P_((CONST char *));
+static u_char *NEAR getentnum __P_((CONST char *, off_t));
 #endif
 static VOID NEAR restorefile __P_((char *, char *, int));
 #endif	/* !_NOWRITEFS */
@@ -100,13 +109,15 @@ static int dosdrv = -1;
 
 
 #ifdef	_USEDOSEMU
-char *nodospath(path, file)
-char *path, *file;
+CONST char *nodospath(path, file)
+char *path;
+CONST char *file;
 {
 	if (!_dospath(file)) return(file);
 	path[0] = '.';
 	path[1] = _SC_;
 	strcpy(&(path[2]), file);
+
 	return(path);
 }
 #endif	/* _USEDOSEMU */
@@ -132,6 +143,7 @@ gid_t gid;
 	else if (isgroupmember(gid)) mode >>= 3;
 #endif	/* !NOUID */
 	if (dir && !(mode & F_ISEXE)) mode &= ~(F_ISRED | F_ISWRI);
+
 	return(mode & 007);
 }
 
@@ -142,7 +154,7 @@ namelist *namep;
 	char path[MAXPATHLEN];
 #endif
 	struct stat st, lst;
-	char *cp;
+	CONST char *cp;
 
 	cp = nodospath(path, namep -> name);
 	if (Xlstat(cp, &lst) < 0 || stat2(cp, &st) < 0) return(-1);
@@ -154,7 +166,8 @@ namelist *namep;
 		memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
 #if	!MSDOS
 	if ((lst.st_mode & S_IFMT) == S_IFCHR
-	|| (lst.st_mode & S_IFMT) == S_IFBLK) namep -> flags |= F_ISDEV;
+	|| (lst.st_mode & S_IFMT) == S_IFBLK)
+		namep -> flags |= F_ISDEV;
 #endif
 
 	namep -> st_mode = lst.st_mode;
@@ -172,14 +185,9 @@ namelist *namep;
 	namep -> st_flags = lst.st_flags;
 #endif
 	namep -> st_mtim = lst.st_mtime;
-	namep -> flags |=
-#ifdef	NOUID
-		logical_access((u_int)(st.st_mode));
-#else
-		logical_access((u_int)(st.st_mode), st.st_uid, st.st_gid);
-#endif
-
+	namep -> flags |= logical_access2(&st);
 	namep -> tmpflags |= F_STAT;
+
 	return(0);
 }
 
@@ -202,13 +210,8 @@ CONST VOID_P vp2;
 
 	if (!isdir(namep1) && isdir(namep2)) return(1);
 	if (isdir(namep1) && !isdir(namep2)) return(-1);
-
-	if (namep1 -> name[0] == '.' && namep1 -> name[1] == '\0') return(-1);
-	if (namep2 -> name[0] == '.' && namep2 -> name[1] == '\0') return(1);
-	if (namep1 -> name[0] == '.' && namep1 -> name[1] == '.'
-	&& namep1 -> name[2] == '\0') return(-1);
-	if (namep2 -> name[0] == '.' && namep2 -> name[1] == '.'
-	&& namep2 -> name[2] == '\0') return(1);
+	if ((tmp = isdotdir(namep2 -> name) - isdotdir(namep1 -> name)))
+		return(tmp);
 
 	switch (sorton & 7) {
 		case 5:
@@ -275,23 +278,23 @@ int cmptree(vp1, vp2)
 CONST VOID_P vp1;
 CONST VOID_P vp2;
 {
-	char *cp1, *cp2;
+	treelist *tp1, *tp2;
+	CONST char *cp1, *cp2;
 	int tmp;
 
-	if (!(((treelist *)vp1) -> name)) return(1);
-	if (!(((treelist *)vp2) -> name)) return(-1);
+	tp1 = (treelist *)vp1;
+	tp2 = (treelist *)vp2;
+	if (!(tp1 -> name)) return(1);
+	if (!(tp2 -> name)) return(-1);
 	switch (sorton & 7) {
 		case 1:
-			tmp = strpathcmp2(((treelist *)vp1) -> name,
-				((treelist *)vp2) -> name);
+			tmp = strpathcmp2(tp1 -> name, tp2 -> name);
 			break;
 		case 2:
-			if ((cp1 = strrchr(((treelist *)vp1) -> name, '.')))
-				cp1++;
-			else cp1 = "";
-			if ((cp2 = strrchr(((treelist *)vp2) -> name, '.')))
-				cp2++;
-			else cp2 = "";
+			if ((cp1 = strrchr(tp1 -> name, '.'))) cp1++;
+			else cp1 = nullstr;
+			if ((cp2 = strrchr(tp2 -> name, '.'))) cp2++;
+			else cp2 = nullstr;
 			tmp = strpathcmp2(cp1, cp2);
 			break;
 		case 3:
@@ -300,17 +303,17 @@ CONST VOID_P vp2;
 			tmp = 0;
 			break;
 	}
-
 	if (sorton > 7) tmp = -tmp;
+
 	return(tmp);
 }
-#endif
+#endif	/* !_NOTREE */
 
 /*ARGSUSED*/
 struct dirent *searchdir(dirp, regexp, arcregstr)
 DIR *dirp;
-reg_t *regexp;
-char *arcregstr;
+CONST reg_t *regexp;
+CONST char *arcregstr;
 {
 	struct dirent *dp;
 #ifndef	_NOARCHIVE
@@ -335,6 +338,7 @@ char *arcregstr;
 	}
 #endif
 	else dp = Xreaddir(dirp);
+
 	return(dp);
 }
 
@@ -342,7 +346,8 @@ int underhome(buf)
 char *buf;
 {
 	static char *homedir = NULL;
-	char *cp, cwd[MAXPATHLEN];
+	CONST char *cp;
+	char tmp[MAXPATHLEN], cwd[MAXPATHLEN];
 	int len;
 
 #if	MSDOS
@@ -354,13 +359,11 @@ char *buf;
 	}
 	if (!homedir) {
 		if (!(cp = gethomedir())) return(-1);
-		if (
 #ifndef	_NODOSDRIVE
-		dospath2(cp) ||
+		if (dospath2(cp)) return(-1);
 #endif
-		_chdir2(cp) < 0) return(-1);
-		homedir = getwd2();
-		if (_chdir2(cwd) < 0) lostcwd(cwd);
+		if (getrealpath(cp, tmp, cwd) != tmp) return(-1);
+		homedir = strdup2(tmp);
 	}
 
 	if (buf && !physical_path) strcpy(cwd, fullpath);
@@ -372,20 +375,21 @@ char *buf;
 #endif
 	else cp = underpath(cwd, homedir, len);
 
-	if (buf) strcpy(buf, (cp) ? cp : "");
+	if (buf) strcpy(buf, (cp) ? cp : nullstr);
 #ifdef	DEBUG
 	free(homedir);
 	homedir = NULL;
 #endif
+
 	return((cp) ? 1 : 0);
 }
 
 int preparedir(dir)
-char *dir;
+CONST char *dir;
 {
 	struct stat st;
 	char tmp[MAXPATHLEN];
-	char *cp;
+	CONST char *cp;
 
 	cp = dir;
 #ifdef	_USEDOSPATH
@@ -413,95 +417,396 @@ char *dir;
 			return(-1);
 		}
 	}
-
 	entryhist(1, realpath2(dir, tmp, 0), 1);
+
 	return(0);
 }
 
+#ifndef	NOFLOCK
+static int NEAR fcntllock(fd, mode)
+int fd, mode;
+{
+	static int lockmode[] = {
+# ifdef	USEFCNTLOCK
+		F_RDLCK, F_WRLCK, F_UNLCK,
+# else	/* !USEFCNTLOCK */
+#  ifdef	USELOCKF
+		F_TLOCK, F_TLOCK, F_ULOCK,
+#  else
+		LOCK_SH | LOCK_NB, LOCK_EX | LOCK_NB, LOCK_UN,
+#  endif
+# endif	/* !USEFCNTLOCK */
+	};
+# ifdef	USEFCNTLOCK
+	struct flock lock;
+# endif
+	int i, n;
+
+# ifndef	_NODOSDRIVE
+	if (fd >= DOSFDOFFSET) return(0);
+# endif
+
+	n = -1;
+	errno = 0;
+	for (i = 0; i < LCK_MAXRETRY; i++) {
+# ifdef	USEFCNTLOCK
+		lock.l_type = lockmode[mode];
+		lock.l_start = lock.l_len = (off_t)0;
+		lock.l_whence = SEEK_SET;
+		n = fcntl(fd, F_SETLK, &lock);
+# else	/* !USEFCNTLOCK */
+#  ifdef	USELOCKF
+		n = lockf(fd, lockmode[mode], (off_t)0);
+#  else
+		n = flock(fd, lockmode[mode]);
+#  endif
+# endif	/* !USEFCNTLOCK */
+		if (n >= 0) {
+			n = 1;
+			break;
+		}
+# ifdef	EACCES
+		else if (errno == EACCES) /*EMPTY*/;
+# endif
+# ifdef	EAGAIN
+		else if (errno == EAGAIN) /*EMPTY*/;
+# endif
+# ifdef	EWOULDBLOCK
+		else if (errno == EWOULDBLOCK) /*EMPTY*/;
+# endif
+		else break;
+
+		if (intrkey(-1)) {
+			errno = EINTR;
+			break;
+		}
+# if	!MSDOS || defined (DJGPP)
+		usleep(100000L);
+# endif
+	}
+	if (i >= LCK_MAXRETRY) return(0);
+
+	return(n);
+}
+#endif	/* !NOFLOCK */
+
+static char *NEAR excllock(file, mode)
+CONST char *file;
+int mode;
+{
+#if	MSDOS
+	char *ext;
+#endif
+	static char **locklist = NULL;
+	struct stat st;
+	char *cp, path[MAXPATHLEN];
+	int i, fd;
+
+	if (!file) {
+		if (locklist) {
+			for (i = 0; locklist[i]; i++) {
+				VOID_C unlink(locklist[i]);
+				free(locklist[i]);
+			}
+			free(locklist);
+		}
+		return(NULL);
+	}
+
+	if (mode == LCK_UNLOCK) {
+		VOID_C Xunlink(file);
+		if (locklist) {
+			for (i = 0; locklist[i]; i++)
+				if (file == locklist[i]) break;
+			if (locklist[i]) {
+				for (; locklist[i + 1]; i++)
+					locklist[i] = locklist[i + 1];
+				locklist[i] = NULL;
+			}
+			if (!i) {
+				free(locklist);
+				locklist = NULL;
+			}
+		}
+		return(NULL);
+	}
+
+#if	MSDOS
+	cp = getbasename(file);
+	for (i = 0; cp[i]; i++) if (cp[i] == '.') break;
+	ext = &(cp[i]);
+	if (i > DOSBODYLEN - strsize(LOCKEXT))
+		i = DOSBODYLEN - strsize(LOCKEXT);
+	i += cp - file;
+	snprintf2(path, sizeof(path), "%-.*s%s%s", i, file, LOCKEXT, ext);
+#else
+	snprintf2(path, sizeof(path), "%s.%s", file, LOCKEXT);
+#endif
+
+	fd = -1;
+	for (i = 0; i < LCK_MAXRETRY; i++) {
+		if (mode != LCK_READ)
+			fd = Xopen(path,
+				O_BINARY | O_WRONLY | O_CREAT | O_EXCL,
+				0666 & ~tmpumask);
+		else if (Xlstat(path, &st) < 0)
+			fd = (errno == ENOENT) ? 0 : -1;
+		else {
+			errno = EEXIST;
+			fd = -1;
+		}
+		if (fd >= 0 || errno != EEXIST) break;
+
+		if (intrkey(-1)) {
+			errno = EINTR;
+			break;
+		}
+#if	!MSDOS || defined (DJGPP)
+		usleep(100000L);
+#endif
+	}
+	if (i >= LCK_MAXRETRY) return(vnullstr);
+	if (fd < 0) return(NULL);
+	if (mode == LCK_READ) return((char *)file);
+
+	Xclose(fd);
+	cp = strdup2(path);
+	i = countvar(locklist);
+	locklist = (char **)realloc2(locklist, (i + 2) * sizeof(char *));
+	locklist[i] = cp;
+	locklist[++i] = NULL;
+
+	return(cp);
+}
+
+lockbuf_t *lockopen(path, flags, mode)
+CONST char *path;
+int flags, mode;
+{
+#ifndef	NOFLOCK
+	int n;
+#endif
+	char *lckname;
+	lockbuf_t *lck;
+	int fd, err, lckflags, lckmode, duperrno;
+
+#ifdef	FAKEUNINIT
+	fd = -1;	/* fake for -Wuninitialized */
+#endif
+	lckname = NULL;
+	lckflags = 0;
+	lckmode = ((flags & O_ACCMODE) == O_RDONLY) ? LCK_READ : LCK_WRITE;
+	err = 0;
+
+#ifndef	NOFLOCK
+	if (isnfs(path) <= 0) {
+		lckflags |= LCK_FLOCK;
+		fd = newdup(Xopen(path, flags & ~O_TRUNC, mode));
+		if (fd < 0) {
+			if ((flags & O_ACCMODE) == O_WRONLY || errno != ENOENT)
+				return(NULL);
+		}
+		else if ((n = fcntllock(fd, lckmode)) <= 0) {
+			Xclose(fd);
+			if (n < 0) return(NULL);
+			err++;
+			lckflags &= ~LCK_FLOCK;
+		}
+		else if ((flags & O_TRUNC) && Xftruncate(fd, (off_t)0) < 0) {
+			duperrno = errno;
+			VOID_C fcntllock(fd, LCK_UNLOCK);
+			Xclose(fd);
+			errno = duperrno;
+			return(NULL);
+		}
+	}
+#endif	/* !NOFLOCK */
+
+	if (!(lckflags & LCK_FLOCK)) {
+		if (!(lckname = excllock(path, lckmode))) return(NULL);
+		else if (lckmode == LCK_READ) {
+			if (lckname == vnullstr) err++;
+			lckname = NULL;
+		}
+		else if (lckname == vnullstr) return(NULL);
+
+		if ((fd = newdup(Xopen(path, flags, mode))) >= 0) /*EMPTY*/;
+		else if ((flags & O_ACCMODE) == O_WRONLY || errno != ENOENT) {
+			duperrno = errno;
+			if (lckname) {
+				VOID_C excllock(lckname, LCK_UNLOCK);
+				free(lckname);
+			}
+			errno = duperrno;
+			return(NULL);
+		}
+	}
+
+	if (fd < 0) lckflags |= LCK_INVALID;
+	else if (err) {
+		fprintf2(stderr, "%k: %k\r\n", path, NOLCK_K);
+		if (isttyiomode) warning(0, HITKY_K);
+	}
+
+	lck = (lockbuf_t *)malloc2(sizeof(lockbuf_t));
+	lck -> fd = fd;
+	lck -> fp = NULL;
+	lck -> name = lckname;
+	lck -> flags = lckflags;
+
+	return(lck);
+}
+
+lockbuf_t *lockfopen(path, type, flags)
+CONST char *path, *type;
+int flags;
+{
+	lockbuf_t *lck;
+	FILE *fp;
+
+	if (!(lck = lockopen(path, flags, 0666))) return(NULL);
+
+	if (!(lck -> flags & LCK_INVALID)) {
+		if (!(fp = Xfdopen(lck -> fd, type))) {
+			lockclose(lck);
+			return(NULL);
+		}
+		lck -> fp = fp;
+	}
+	lck -> flags |= LCK_STREAM;
+
+	return(lck);
+}
+
+VOID lockclose(lck)
+lockbuf_t *lck;
+{
+	if (lck) {
+		if (lck -> name) {
+			VOID_C excllock(lck -> name, LCK_UNLOCK);
+			free(lck -> name);
+		}
+		if (!(lck -> flags & LCK_INVALID)) {
+#ifndef	NOFLOCK
+			if (lck -> flags & LCK_FLOCK)
+				VOID_C fcntllock(lck -> fd, LCK_UNLOCK);
+#endif
+			if (lck -> flags & LCK_STREAM) Xfclose(lck -> fp);
+			else Xclose(lck -> fd);
+		}
+
+		free(lck);
+	}
+}
+
 int touchfile(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
-	struct stat st, dummy;
+#ifdef	USEUTIME
+	struct utimbuf times;
+#else
+	struct timeval tvp[2];
+#endif
+#ifndef	_NOEXTRAATTR
+	int i;
+#endif
+	struct stat st;
+	u_int mode;
 	int ret, duperrno;
 
+	if (!(stp -> st_nlink)) return(0);
 	if (Xlstat(path, &st) < 0) return(-1);
 	if (s_islnk(&st)) return(1);
 
 	ret = 0;
-#ifdef	FAKEUNINIT
-	duperrno = errno;	/* fake for -Wuninitialized */
-#endif
 #if	MSDOS
-	if (!(st.st_mode & S_IWRITE)) Xchmod(path, (st.st_mode | S_IWRITE));
+	if (!(st.st_mode & S_IWUSR)) Xchmod(path, (st.st_mode | S_IWUSR));
 #endif
-	if (stp -> st_atime != (time_t)-1 && stp -> st_mtime != (time_t)-1) {
-#ifdef	USEUTIME
-		struct utimbuf times;
 
+	if (stp -> st_nlink & (TCH_ATIME | TCH_MTIME)) {
+		if (!(stp -> st_nlink & TCH_ATIME))
+			stp -> st_atime = st.st_atime;
+		if (!(stp -> st_nlink & TCH_MTIME))
+			stp -> st_mtime = st.st_mtime;
+#ifdef	USEUTIME
 		times.actime = stp -> st_atime;
 		times.modtime = stp -> st_mtime;
-		if (Xutime(path, &times) < 0) {
-			duperrno = errno;
-			ret = -1;
-		}
+		duperrno = errno;
+		if (Xutime(path, &times) >= 0) errno = duperrno;
+		else ret = -1;
 #else
-		struct timeval tvp[2];
-
 		tvp[0].tv_sec = stp -> st_atime;
 		tvp[0].tv_usec = 0;
 		tvp[1].tv_sec = stp -> st_mtime;
 		tvp[1].tv_usec = 0;
-		if (Xutimes(path, tvp) < 0) {
-			duperrno = errno;
-			ret = -1;
-		}
+		duperrno = errno;
+		if (Xutimes(path, tvp) >= 0) errno = duperrno;
+		else ret = -1;
 #endif
 	}
+
 #ifdef	HAVEFLAGS
-	if (stp -> st_flags != (u_long)-1) {
-# ifndef	_NODOSDRIVE
-		if (dospath(path, NULL)) {
-			duperrno = EACCES;
-			ret = -1;
-		}
-		else
-# endif
-		if (chflags(path, stp -> st_flags) < 0) {
-			duperrno = errno;
-			ret = -1;
-		}
+	if (stp -> st_nlink & TCH_FLAGS) {
+		duperrno = errno;
+		if (Xchflags(path, stp -> st_flags) >= 0) errno = duperrno;
+		else ret = -1;
 	}
 #endif
-	dummy.st_mode = (u_short)-1;
-	if (stp -> st_mode != dummy.st_mode) {
-		stp -> st_mode &= ~S_IFMT;
-		stp -> st_mode |= (st.st_mode & S_IFMT);
-		if (Xchmod(path, stp -> st_mode) < 0) {
-			duperrno = errno;
-			ret = -1;
+
+	if (stp -> st_nlink & TCH_MODE) {
+		mode = stp -> st_mode;
+		mode &= ~S_IFMT;
+		mode |= (st.st_mode & S_IFMT);
+#ifndef	_NOEXTRAATTR
+		if ((stp -> st_nlink & TCH_MODEEXE)
+		&& !s_isdir(&st) && !(st.st_mode & S_IEXEC_ALL)) {
+			for (i = 0; i < 3; i++) {
+				if (!(stp -> st_mode & (1 << (i + 12))))
+					continue;
+				mode &= ~(1 << (i * 3));
+				mode |= (st.st_mode & (1 << (i * 3)));
+			}
 		}
+		if (stp -> st_nlink & TCH_MASK) {
+			mode &= ~(stp -> st_size);
+			mode |= (st.st_mode & (stp -> st_size));
+		}
+#endif	/* !_NOEXTRAATTR */
+		duperrno = errno;
+		if (Xchmod(path, mode) >= 0) errno = duperrno;
+		else ret = -1;
 	}
 #if	MSDOS
-	else if (!(st.st_mode & S_IWRITE)) Xchmod(path, st.st_mode);
+	else if (!(st.st_mode & S_IWUSR)) {
+		duperrno = errno;
+		Xchmod(path, st.st_mode);
+		errno = duperrno;
+	}
 #endif
+
 #ifndef	NOUID
-	if (stp -> st_gid != (gid_t)-1) {
-# ifndef	_NODOSDRIVE
-		if (dospath(path, NULL)) /*EMPTY*/;
-		else
-# endif
-		if (chown(path, stp -> st_uid, stp -> st_gid) < 0)
-			chown(path, (uid_t)-1, stp -> st_gid);
+	if (stp -> st_nlink & (TCH_UID | TCH_GID)) {
+		if (!(stp -> st_nlink & TCH_UID)) stp -> st_uid = (uid_t)-1;
+		if (!(stp -> st_nlink & TCH_GID)) stp -> st_gid = (gid_t)-1;
+		duperrno = errno;
+		if (Xchown(path, stp -> st_uid, stp -> st_gid) >= 0)
+			errno = duperrno;
+		else if (!(stp -> st_nlink & TCH_CHANGE)) {
+			Xchown(path, (uid_t)-1, stp -> st_gid);
+			errno = duperrno;
+		}
+		else ret = -1;
 	}
 #endif	/* !NOUID */
-	if (ret < 0) errno = duperrno;
+	if (stp -> st_nlink & TCH_IGNOREERR) ret = 0;
+
 	return(ret);
 }
 
 #ifndef	_NODOSDRIVE
 static int NEAR nodoschdir(path)
-char *path;
+CONST char *path;
 {
 	int n, dupdosdrive;
 
@@ -509,6 +814,7 @@ char *path;
 	dosdrive = 0;
 	n = Xchdir(path);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -522,11 +828,12 @@ char *path;
 	dosdrive = 0;
 	cp = Xgetwd(path);
 	dosdrive = dupdosdrive;
+
 	return(cp);
 }
 
 int nodoslstat(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 	int n, dupdosdrive;
@@ -535,11 +842,12 @@ struct stat *stp;
 	dosdrive = 0;
 	n = Xlstat(path, stp);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
 static int NEAR nodosmkdir(path, mode)
-char *path;
+CONST char *path;
 int mode;
 {
 	int n, dupdosdrive;
@@ -548,11 +856,12 @@ int mode;
 	dosdrive = 0;
 	n = Xmkdir(path, mode);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
 static int NEAR nodosrmdir(path)
-char *path;
+CONST char *path;
 {
 	int n, dupdosdrive;
 
@@ -560,6 +869,7 @@ char *path;
 	dosdrive = 0;
 	n = Xrmdir(path);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 #endif	/* !_NODOSDRIVE */
@@ -567,111 +877,157 @@ char *path;
 VOID lostcwd(path)
 char *path;
 {
-	char *cp, buf[MAXPATHLEN];
+	CONST char *cp;
+	char buf[MAXPATHLEN];
 	int duperrno;
 
 	duperrno = errno;
 	if (!path) path = buf;
 
-	if (path != fullpath && !nodoschdir(fullpath) && nodosgetwd(path))
+	if (path != fullpath && nodoschdir(fullpath) >= 0 && nodosgetwd(path))
 		cp = NOCWD_K;
-	else if ((cp = gethomedir()) && !nodoschdir(cp) && nodosgetwd(path))
+	else if ((cp = gethomedir())
+	&& nodoschdir(cp) >= 0 && nodosgetwd(path))
 		cp = GOHOM_K;
-	else if (!nodoschdir(_SS_) && nodosgetwd(path)) cp = GOROT_K;
-	else error(_SS_);
+	else if (nodoschdir(rootpath) >= 0 && nodosgetwd(path)) cp = GOROT_K;
+	else error(rootpath);
 
 	if (path != fullpath) strncpy2(fullpath, path, MAXPATHLEN - 1);
 	warning(0, cp);
 #ifndef	_NOUSEHASH
-	searchhash(NULL, "", "");
+	searchhash(NULL, nullstr, nullstr);
 #endif
 	errno = duperrno;
 }
 
-int safewrite(fd, buf, size)
-int fd;
-char *buf;
-int size;
+#ifndef	NODIRLOOP
+int issamebody(src, dest)
+CONST char *src, *dest;
 {
-	int n;
+	struct stat st1, st2;
 
-	n = Xwrite(fd, buf, size);
-#if	MSDOS
-	if (n >= 0 && n < size) {
-		n = -1;
-		errno = ENOSPC;
-	}
-#else
-	if (n < 0 && errno == EINTR) n = size;
-#endif
-	return(n);
+	if (Xstat(src, &st1) < 0 || Xstat(dest, &st2) < 0) return(0);
+	if (st1.st_dev != st2.st_dev || st1.st_ino != st2.st_ino) return(0);
+
+	return(1);
 }
+#endif	/* !NODIRLOOP */
+
+#ifndef	NOSYMLINK
+int cpsymlink(src, dest)
+CONST char *src, *dest;
+{
+	struct stat st;
+	char path[MAXPATHLEN];
+	int len;
+
+	if ((len = Xreadlink(src, path, strsize(path))) < 0) return(-1);
+	if (Xlstat(dest, &st) >= 0) {
+# ifndef	NODIRLOOP
+		if (issamebody(src, dest)) return(0);
+# endif
+		if (Xunlink(dest) < 0) return(-1);
+	}
+	path[len] = '\0';
+	if (Xsymlink(path, dest) < 0) return(-1);
+
+	return(1);
+}
+#endif	/* !NOSYMLINK */
 
 /*ARGSUSED*/
 static int NEAR cpfile(src, dest, stp1, stp2)
-char *src, *dest;
+CONST char *src, *dest;
 struct stat *stp1, *stp2;
 {
 #if	MSDOS
 	struct stat st;
+#endif
+#ifndef	NODIRLOOP
+	int issame;
 #endif
 	char buf[BUFSIZ];
-	int i, fd1, fd2, duperrno;
+	off_t total;
+	int n, fd1, fd2, tty, flags, mode, duperrno;
 
 #if	MSDOS
 	if (!stp2 && Xlstat(dest, &st) >= 0) stp2 = &st;
-	if (stp2 && !(stp2 -> st_mode & S_IWRITE))
-		Xchmod(src, stp2 -> st_mode | S_IWRITE);
+	if (stp2 && !(stp2 -> st_mode & S_IWUSR))
+		Xchmod(src, stp2 -> st_mode | S_IWUSR);
 #endif
+#ifndef	NOSYMLINK
 	if (s_islnk(stp1)) {
-		if ((i = Xreadlink(src, buf, BUFSIZ - 1)) < 0) return(-1);
-		buf[i] = '\0';
-		return(Xsymlink(buf, dest));
+		if (cpsymlink(src, dest) < 0) return(-1);
+		else return(0);
 	}
-	if ((fd1 = Xopen(src, O_BINARY | O_RDONLY, stp1 -> st_mode)) < 0)
-		return(-1);
-#if	MSDOS
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC,
-		stp1 -> st_mode | S_IWRITE);
-#else
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC, stp1 -> st_mode);
 #endif
-	if (fd2 < 0) {
-		Xclose(fd1);
-		return(-1);
-	}
 
-	for (;;) {
-		while ((i = Xread(fd1, buf, BUFSIZ)) < 0 && errno == EINTR);
-		duperrno = errno;
-		if (i < BUFSIZ) break;
-		if ((i = safewrite(fd2, buf, BUFSIZ)) < 0) {
-			duperrno = errno;
-			break;
+	flags = (O_BINARY | O_RDONLY);
+	mode = stp1 -> st_mode;
+#ifndef	NODIRLOOP
+	issame = issamebody(src, dest);
+#endif
+	if ((fd1 = Xopen(src, flags, mode)) < 0) return(-1);
+
+	flags = (O_BINARY | O_WRONLY | O_CREAT | O_TRUNC);
+#if	MSDOS
+	mode |= S_IWUSR;
+#endif
+#ifndef	NODIRLOOP
+	if (issame) {
+		flags |= O_EXCL;
+		if (Xunlink(dest) < 0) {
+			VOID_C Xclose(fd1);
+			return(-1);
 		}
 	}
-	if (i > 0 && (i = safewrite(fd2, buf, i)) < 0) duperrno = errno;
+#endif
+	if ((fd2 = Xopen(dest, flags, mode)) < 0) {
+		VOID_C Xclose(fd1);
+		return(-1);
+	}
 
-	Xclose(fd2);
-	Xclose(fd1);
+	tty = isatty(fd1);
+	total = (off_t)0;
+	for (;;) {
+		if (!tty && total >= stp1 -> st_size) n = 0;
+		else n = sureread(fd1, buf, sizeof(buf));
+		if (n < 0) break;
+		total += (off_t)n;
+		if (!n) {
+			if (!tty && total < stp1 -> st_size) {
+				duperrno = ETIMEDOUT;
+				n = -1;
+			}
+			break;
+		}
+#ifndef	_NOEXTRACOPY
+		showprogress(n);
+#endif
+		if ((n = surewrite(fd2, buf, n)) < 0) break;
+	}
 
-	if (i < 0) {
-		Xunlink(dest);
+	safeclose(fd2);
+	safeclose(fd1);
+#ifndef	_NOEXTRACOPY
+	fshowprogress(dest);
+#endif
+
+	if (n < 0) {
+		duperrno = errno;
+		VOID_C Xunlink(dest);
 		errno = duperrno;
 		return(-1);
 	}
-#ifdef	HAVEFLAGS
-	stp1 -> st_flags = (u_long)-1;
-#endif
-#ifdef	_USEDOSCOPY
-	if (touchfile(dest, stp1) < 0) return(-1);
-#else
-# ifndef	_NOEXTRACOPY
-	if (inheritcopy && touchfile(dest, stp1) < 0) return(-1);
+	stp1 -> st_nlink = (TCH_ATIME | TCH_MTIME | TCH_IGNOREERR);
+#if	defined (_USEDOSCOPY) || !defined (_NOEXTRACOPY)
+# ifndef	_USEDOSCOPY
+	if (!inheritcopy) /*EMPTY*/;
+	else
 # endif
-#endif
+	if (touchfile(dest, stp1) < 0) return(-1);
+#endif	/* _USEDOSCOPY || !_NOEXTRACOPY */
+
 	return(0);
 }
 
@@ -681,7 +1037,7 @@ static VOID changemes(VOID_A)
 }
 
 int safecpfile(src, dest, stp1, stp2)
-char *src, *dest;
+CONST char *src, *dest;
 struct stat *stp1, *stp2;
 {
 #ifndef	_NODOSDRIVE
@@ -711,16 +1067,21 @@ struct stat *stp1, *stp2;
 }
 
 int safemvfile(src, dest, stp1, stp2)
-char *src, *dest;
+CONST char *src, *dest;
 struct stat *stp1, *stp2;
 {
-	if (Xrename(src, dest) >= 0) return(0);
+	if (Xrename(src, dest) >= 0) {
+#ifndef	_NOEXTRACOPY
+		fshowprogress(dest);
+#endif
+		return(0);
+	}
 	if (errno != EXDEV || s_isdir(stp1)) return(-1);
 	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
 		return(-1);
-#ifdef	HAVEFLAGS
-	stp1 -> st_flags = (u_long)-1;
-#endif
+
+	stp1 -> st_nlink = (TCH_MODE | TCH_UID | TCH_GID
+		| TCH_ATIME | TCH_MTIME | TCH_IGNOREERR);
 	return (touchfile(dest, stp1));
 }
 
@@ -762,8 +1123,8 @@ int len;
 	int i, j, c;
 
 	if (!buf) {
-		for (i = 0; i < sizeof(seq) / sizeof(char); i++) {
-			j = genrand(sizeof(seq) / sizeof(char));
+		for (i = 0; i < arraysize(seq); i++) {
+			j = genrand(arraysize(seq));
 			c = seq[i];
 			seq[i] = seq[j];
 			seq[j] = c;
@@ -771,11 +1132,12 @@ int len;
 	}
 	else {
 		for (i = 0; i < len; i++) {
-			j = genrand(sizeof(seq) / sizeof(char));
+			j = genrand(arraysize(seq));
 			buf[i] = seq[j];
 		}
 		buf[i] = '\0';
 	}
+
 	return(buf);
 }
 
@@ -800,8 +1162,7 @@ char *dir;
 #endif
 	deftmpdir = strdup2(path);
 #if	MSDOS
-	n = getcurdrv();
-	*path = (isupper2(n)) ? toupper2(*path) : tolower2(*path);
+	*path = (isupper2(getcurdrv())) ? toupper2(*path) : tolower2(*path);
 #endif
 
 	mask = 0777 & ~tmpumask;
@@ -814,9 +1175,9 @@ char *dir;
 		}
 	}
 	if (!tmpfilename) {
-		n = sizeof(TMPPREFIX) - 1;
+		n = strsize(TMPPREFIX);
 		strncpy(cp, TMPPREFIX, n);
-		len = sizeof(path) - 1 - (cp - path);
+		len = strsize(path) - (cp - path);
 		if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 		len -= n;
 		genrandname(NULL, 0);
@@ -835,7 +1196,7 @@ char *dir;
 	}
 
 	strncpy((cp = strcatdelim(path)), dir, n);
-	len = sizeof(path) - 1 - (cp - path);
+	len = strsize(path) - (cp - path);
 	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 	len -= n;
 	genrandname(NULL, 0);
@@ -855,11 +1216,12 @@ char *dir;
 		nodosrmdir(path);
 		errno = n;
 	}
+
 	return(-1);
 }
 
 int rmtmpdir(dir)
-char *dir;
+CONST char *dir;
 {
 	char path[MAXPATHLEN];
 
@@ -875,61 +1237,81 @@ char *dir;
 	}
 	else if (errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
 		return(-1);
+
 	return(0);
 }
 
-int mktmpfile(file)
-char *file;
+int opentmpfile(path, mode)
+CONST char *path;
+int mode;
 {
-	char *cp, path[MAXPATHLEN];
-	int fd, len, duperrno;
+	char *cp;
+	int fd, len;
 
-	path[0] = '\0';
-	if (mktmpdir(path) < 0) return(-1);
-	cp = strcatdelim(path);
-	len = sizeof(path) - 1 - (cp - path);
+	cp = getbasename(path);
+	len = MAXPATHLEN - (cp - path) - 1;
 	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 	genrandname(NULL, 0);
 
 	for (;;) {
 		genrandname(cp, len);
-		fd = Xopen(path, O_BINARY | O_WRONLY | O_CREAT | O_EXCL,
-			0666 & ~tmpumask);
-		if (fd >= 0) {
-			strcpy(file, path);
-			return(fd);
-		}
-		if (errno != EEXIST) break;
+		fd = Xopen(path, O_BINARY | O_WRONLY | O_CREAT | O_EXCL, mode);
+		if (fd >= 0) break;
+		if (errno != EEXIST) return(-1);
+	}
+
+	return(fd);
+}
+
+int mktmpfile(file)
+char *file;
+{
+	char path[MAXPATHLEN];
+	int fd, duperrno;
+
+	path[0] = '\0';
+	if (mktmpdir(path) < 0) return(-1);
+	VOID_C strcatdelim(path);
+	if ((fd = opentmpfile(path, 0666 & ~tmpumask)) >= 0) {
+		strcpy(file, path);
+		return(fd);
 	}
 	duperrno = errno;
 	rmtmpdir(NULL);
 	errno = duperrno;
+
 	return(-1);
 }
 
 int rmtmpfile(file)
-char *file;
+CONST char *file;
 {
 	if ((Xunlink(file) < 0 && errno != ENOENT) || rmtmpdir(NULL) < 0)
 		return(-1);
+
 	return(0);
 }
 
 static int dounlink(path)
-char *path;
+CONST char *path;
 {
-	return(Xunlink(path));
+	if (Xunlink(path) < 0) return(APL_ERROR);
+
+	return(APL_OK);
 }
 
 static int dormdir(path)
-char *path;
+CONST char *path;
 {
-	if (isdotdir(path)) return(0);
-	return(Xrmdir(path));
+	if (isdotdir(path)) return(APL_OK);
+	if (Xrmdir(path) < 0) return(APL_ERROR);
+
+	return(APL_OK);
 }
 
 VOID removetmp(dir, file)
-char *dir, *file;
+char *dir;
+CONST char *file;
 {
 #ifdef	_USEDOSEMU
 	char path[MAXPATHLEN];
@@ -940,24 +1322,15 @@ char *dir, *file;
 		warning(-1, dir);
 		*dir = '\0';
 	}
-	else if (!file) applydir(NULL, dounlink, NULL, dormdir, 3, NULL);
+	else if (!file) VOID_C applydir(NULL, dounlink,
+		NULL, dormdir, ORD_NOPREDIR, NULL);
 	else if (Xunlink(nodospath(path, file)) < 0) warning(-1, file);
 
-#ifndef	_NODOSDRIVE
-	if (dosdrv >= 0) {
-		if (lastdrv >= 0) {
-# if	MSDOS
-			if (_chdir2(deftmpdir) < 0);
-			else
-# endif
-			shutdrv(lastdrv);
-		}
-		lastdrv = dosdrv;
-		dosdrv = -1;
-	}
-#endif	/* !_NODOSDRIVE */
-
 	if (_chdir2(fullpath) < 0) lostcwd(fullpath);
+#ifndef	_NODOSDRIVE
+	shutdrv(dosdrv);
+	dosdrv = -1;
+#endif
 	if (dir) {
 		if (*dir && rmtmpdir(dir) < 0) warning(-1, dir);
 		free(dir);
@@ -965,7 +1338,7 @@ char *dir, *file;
 }
 
 int forcecleandir(dir, file)
-char *dir, *file;
+CONST char *dir, *file;
 {
 #if	!MSDOS
 	p_id_t pid;
@@ -973,11 +1346,13 @@ char *dir, *file;
 	extern char **environ;
 	char buf[MAXPATHLEN];
 
+	VOID_C excllock(NULL, LCK_UNLOCK);
+
 	if (!dir || !*dir || !file || !*file) return(0);
 	strcatdelim2(buf, dir, file);
 
 	if (rawchdir(buf) != 0) return(0);
-	rawchdir(_SS_);
+	rawchdir(rootpath);
 #if	MSDOS
 	spawnlpe(P_WAIT, "DELTREE.EXE", "DELTREE", "/Y", buf, NULL, environ);
 #else
@@ -987,6 +1362,7 @@ char *dir, *file;
 		_exit(1);
 	}
 #endif
+
 	return(0);
 }
 
@@ -1000,18 +1376,20 @@ int drive;
 	path[1] = drive;
 	path[2] = '\0';
 	if (mktmpdir(path) < 0) return(NULL);
+
 	return(strdup2(path));
 }
 
 int tmpdosdupl(dir, dirp, single)
-char *dir, **dirp;
+CONST char *dir;
+char **dirp;
 int single;
 {
-	struct stat st;
-	char *cp, *tmpdir, path[MAXPATHLEN];
-# if	!MSDOS
+# ifdef	_USEDOSEMU
 	char tmp[MAXPATHLEN];
 # endif
+	struct stat st;
+	char *cp, *tmpdir, path[MAXPATHLEN];
 	int i, drive;
 
 	if (!(drive = dospath2(dir))) return(0);
@@ -1028,6 +1406,7 @@ int single;
 		strcpy(cp, filelist[filepos].name);
 		st.st_mode = filelist[filepos].st_mode;
 		st.st_atime = st.st_mtime = filelist[filepos].st_mtim;
+		st.st_size = filelist[filepos].st_size;
 		if (cpfile(fnodospath(tmp, filepos), path, &st, NULL) < 0) {
 			warning(-1, path);
 			removetmp(tmpdir, NULL);
@@ -1039,6 +1418,7 @@ int single;
 		strcpy(cp, filelist[i].name);
 		st.st_mode = filelist[i].st_mode;
 		st.st_atime = st.st_mtime = filelist[i].st_mtim;
+		st.st_size = filelist[i].st_size;
 		if (cpfile(fnodospath(tmp, i), path, &st, NULL) < 0) {
 			warning(-1, path);
 			removetmp(tmpdir, NULL);
@@ -1046,31 +1426,33 @@ int single;
 		}
 	}
 
-	dosdrv = lastdrv;
-	lastdrv = -1;
+	if (preparedrv(dospath(fullpath, NULL), &dosdrv) < 0) {
+		warning(-1, fullpath);
+		removetmp(tmpdir, NULL);
+		return(-1);
+	}
 	if (_chdir2(tmpdir) < 0) {
 		warning(-1, tmpdir);
 		removetmp(tmpdir, NULL);
 		return(-1);
 	}
 	*dirp = tmpdir;
+
 	return(drive);
 }
 
 int tmpdosrestore(drive, file)
 int drive;
-char *file;
+CONST char *file;
 {
 	struct stat st;
 	char path[MAXPATHLEN];
-# if	!MSDOS
+# ifdef	_USEDOSEMU
 	char tmp[MAXPATHLEN];
 # endif
 
-	path[0] = drive;
-	path[1] = ':';
-	strcpy(&(path[2]), file);
-# if	!MSDOS
+	strcpy(gendospath(path, drive, '\0'), file);
+# ifdef	_USEDOSEMU
 	file = nodospath(tmp, file);
 # endif
 	waitmes();
@@ -1083,16 +1465,17 @@ char *file;
 
 #ifndef	_NOWRITEFS
 static int NEAR isexist(file)
-char *file;
+CONST char *file;
 {
 	struct stat st;
 
 	if (Xlstat(file, &st) < 0 && errno == ENOENT) return(0);
+
 	return(1);
 }
 
 static int NEAR realdirsiz(s, dos, boundary, dirsize, ofs)
-char *s;
+CONST char *s;
 int dos, boundary, dirsize, ofs;
 {
 	int i, len, lfn, dot;
@@ -1125,6 +1508,7 @@ int dos, boundary, dirsize, ofs;
 		len += ofs;
 		if (lfn) return((len / LFNENTSIZ + 1) * DOSDIRENT + DOSDIRENT);
 	}
+
 	return(DOSDIRENT);
 }
 
@@ -1140,11 +1524,12 @@ int size, dos, boundary, dirsize, ofs;
 		size -= DOSDIRENT;
 		return((size / DOSDIRENT) * LFNENTSIZ - 1);
 	}
+
 	return(DOSBODYLEN);
 }
 
 static int NEAR saferename(from, to)
-char *from, *to;
+CONST char *from, *to;
 {
 #ifdef	_USEDOSEMU
 	char fpath[MAXPATHLEN], tpath[MAXPATHLEN];
@@ -1155,14 +1540,18 @@ char *from, *to;
 	from = nodospath(fpath, from);
 	to = nodospath(tpath, to);
 #endif
+
 	return(Xrename(from, to));
 }
 
 /*ARGSUSED*/
 static char *NEAR maketmpfile(len, dos, tmpdir, old)
 int len, dos;
-char *tmpdir, *old;
+CONST char *tmpdir, *old;
 {
+#ifndef	PATHNOCASE
+	int i;
+#endif
 	char *fname, path[MAXPATHLEN];
 	int l, fd;
 
@@ -1178,12 +1567,8 @@ char *tmpdir, *old;
 	for (;;) {
 		genrandname(fname, len);
 #ifndef	PATHNOCASE
-		if (dos) {
-			int i;
-
-			for (i = 0; fname[i]; i++)
-				fname[i] = toupper2(fname[i]);
-		}
+		if (dos) for (i = 0; fname[i]; i++)
+			fname[i] = toupper2(fname[i]);
 #endif
 		if (tmpdir) {
 			strcpy(&(path[l]), fname);
@@ -1201,9 +1586,11 @@ char *tmpdir, *old;
 					else break;
 				}
 				else {
-					if ((fd = Xopen(fname,
-					O_CREAT | O_EXCL, 0600))) {
-						Xclose(fd);
+					fd = Xopen(fname,
+						O_WRONLY | O_CREAT | O_EXCL,
+						0600);
+					if (fd >= 0) {
+						VOID_C Xclose(fd);
 						return(fname);
 					}
 					if (errno != EEXIST) break;
@@ -1216,26 +1603,28 @@ char *tmpdir, *old;
 		}
 	}
 	free(fname);
+
 	return(NULL);
 }
 
 #if	!MSDOS
 static off_t NEAR getdirblocksize(dir)
-char *dir;
+CONST char *dir;
 {
 	struct stat st;
 
 	if (Xlstat(dir, &st) < 0) return(getblocksize(dir));
+
 	return((off_t)st.st_size);
 }
 
 static u_char *NEAR getentnum(dir, bsiz)
-char *dir;
+CONST char *dir;
 off_t bsiz;
 {
 	struct stat st;
 	u_char ch, *tmp;
-	int i, c, n, fd;
+	int i, n, fd;
 
 	if (Xlstat(dir, &st) < 0
 	|| (fd = Xopen(dir, O_BINARY | O_RDONLY, 0666)) < 0)
@@ -1248,16 +1637,14 @@ off_t bsiz;
 			free(tmp);
 			return(NULL);
 		}
-		while ((c = Xread(fd, (char *)&ch, sizeof(ch))) < 0
-		&& errno == EINTR);
-		if (c < 0) {
+		if (sureread(fd, &ch, sizeof(ch)) < 0) {
 			free(tmp);
 			return(NULL);
 		}
 		tmp[i] = ch + 1;
 	}
 	tmp[i] = 0;
-	Xclose(fd);
+	VOID_C Xclose(fd);
 
 	return(tmp);
 }
@@ -1299,7 +1686,8 @@ int fs;
 	struct dirent *dp;
 	int dos, boundary, dirsize, namofs;
 	int i, top, size, len, fnamp, ent;
-	char *cp, *tmpdir, **fnamelist, path[MAXPATHLEN];
+	CONST char *cp;
+	char *tmp, *tmpdir, **fnamelist, path[MAXPATHLEN];
 
 	switch (fs) {
 #if	!MSDOS
@@ -1398,7 +1786,7 @@ int fs;
 	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	waitmes();
 
-	if (!(dirp = Xopendir("."))) {
+	if (!(dirp = Xopendir(curpath))) {
 		freevar(fnamelist);
 		noconv--;
 		lostcwd(path);
@@ -1410,12 +1798,12 @@ int fs;
 		else if (!strpathcmp(dp -> d_name, tmpdir)) {
 #if	MSDOS
 			if (!(dp -> d_alias[0])) len = DOSBODYLEN;
-#else
+#else	/* !MSDOS */
 # ifndef	_NODOSDRIVE
-			if (dos == 3 && dp -> d_reclen == DOSDIRENT)
+			if (dos == 3 && wrap_reclen(dp) == DOSDIRENT)
 				len = DOSBODYLEN;
 # endif
-#endif
+#endif	/* !MSDOS */
 			ent = i;
 		}
 		else {
@@ -1434,7 +1822,7 @@ int fs;
 	Xclosedir(dirp);
 
 	if (ent > 0) {
-		if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir))) {
+		if (!(tmp = maketmpfile(len, dos, tmpdir, tmpdir))) {
 			warning(-1, tmpdir);
 			restorefile(tmpdir, path, fnamp);
 			freevar(fnamelist);
@@ -1442,14 +1830,14 @@ int fs;
 			return;
 		}
 		free(tmpdir);
-		tmpdir = cp;
+		tmpdir = tmp;
 		fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	}
 
 #if	!MSDOS
 	if (fs != FSID_EFS) entnum = NULL;	/* except IRIX File System */
-	else if (!(entnum = getentnum(".", persec))) {
-		warning(-1, ".");
+	else if (!(entnum = getentnum(curpath, persec))) {
+		warning(-1, curpath);
 		restorefile(tmpdir, path, fnamp);
 		freevar(fnamelist);
 		noconv--;
@@ -1457,8 +1845,8 @@ int fs;
 	}
 	totalent = headbyte
 		+ realdirsiz(tmpdir, dos, boundary, dirsize, namofs)
-		+ realdirsiz(".", dos, boundary, dirsize, namofs)
-		+ realdirsiz("..", dos, boundary, dirsize, namofs);
+		+ realdirsiz(curpath, dos, boundary, dirsize, namofs)
+		+ realdirsiz(parentpath, dos, boundary, dirsize, namofs);
 	block = tmpno = 0;
 	ptr = 3;
 	totalptr = 0;
@@ -1518,12 +1906,14 @@ int fs;
 	}
 #endif
 
-	if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir))) warning(-1, tmpdir);
+	if (!(tmp = maketmpfile(len, dos, tmpdir, tmpdir)))
+		warning(-1, tmpdir);
 	else {
 		free(tmpdir);
-		tmpdir = cp;
+		tmpdir = tmp;
 	}
-	fnamp = strcatdelim2(path, tmpdir, fnamelist[top]) - path;
+	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
+	snprintf2(&(path[fnamp]), sizeof(path) - fnamp, fnamelist[top]);
 	if (saferename(path, fnamelist[top]) < 0) warning(-1, path);
 	restorefile(tmpdir, path, fnamp);
 
